@@ -15,6 +15,7 @@ Checks that variables are only assigned once.
 import sys
 import pystrict3lib
 import argparse
+import logging
 
 """
 BSD 2-Clause License
@@ -57,7 +58,7 @@ if __name__ == '__main__':
 
     parser = HelpfulParser(
         description=__doc__,
-        epilog="""Johannes Buchner (C) 2020 <johannes.buchner.acad@gmx.com>""",
+        epilog="""Johannes Buchner (C) 2020-2022 <johannes.buchner.acad@gmx.com>""",
         formatter_class=argparse.ArgumentDefaultsHelpFormatter)
 
     parser.add_argument(
@@ -69,9 +70,25 @@ if __name__ == '__main__':
         help="""Also load any modules specified in import statements to check function signatures.
         Warning: can execute arbitrary module code.""")
 
+    parser.add_argument(
+        '--allow-redefining', default=False, action='store_true',
+        help="""Allow redefining variables.""")
+
+    parser.add_argument(
+        '--verbose', '-v', default=False, action='store_true',
+        help="""More verbose logging output.""")
+
     parser.add_argument('filenames', type=str, nargs='+', help="""python files to parse""")
 
     args = parser.parse_args()
 
+    logger = logging.getLogger('pystrict3')
+    handler = logging.StreamHandler(sys.stderr)
+    handler.setLevel(logging.DEBUG if args.verbose else logging.INFO)
+    formatter = logging.Formatter('[%(levelname)s] %(message)s')
+    handler.setFormatter(formatter)
+    logger.addHandler(handler)
+    logger.setLevel(logging.DEBUG if args.verbose else logging.INFO)
+
     module_load_policy = 'all' if args.import_any else 'builtins' if args.import_builtin else 'none'
-    pystrict3lib.main(args.filenames, module_load_policy)
+    pystrict3lib.main(args.filenames, module_load_policy, args.allow_redefining)
